unit xListDecoration;

{ Defines a class of decoration for displaying a simple scrollable list of items; this }
{ class uses the "Macintosh List Manager."  Each entry in the list is a on-line piece  }
{ of text.  You can add and delete entries.   The list has a scroll bar, and the user can }
{ scroll through the list.  The user can select an item in the list by clicking on it.  }
{ (You can turn off this behaviour if you want.)  When the user double-clicks on an item, }
{ a method called HandleDoubleClick is called; you can override this method if you want }
{ to respond to a double click. }

interface

uses
	xWindow;

type
	xList = object(xWindowDecoration)
       { A scrollable list of items; you can add and delete items and user can scroll }
       { through the entries and select an item (this selection feature can be disabled). }

			theList: ListHandle;       { Macintosh List Handler data structure }
			noUserSelect: boolean;   { true if user is not allowed to select or double-click items }
			saveFont: integer;          { font when list was SetUp; all entries are displayed in this font }
			saveSize: integer;          { font size when list was SetUp }
			saveFace: style;             { font style (bold, italic, etc.) when list was SetUp }

			procedure SetUp (win: xWindow;
										initialEntries: string;
										theLeft, theRight, fixedWidth, theHeight: integer);
         { Creates and intalls a list decoration in the specified window.  InitialEntries }
         { specifies to be inserted into the list initially; this can consist of a number of }
         { entries, with individual entries separated by backslashes ('\').  The remaining }
         { specify the location and size of the list; see the comment on procedure }
         { xWindowDecoration.Install in file xWindow.p for information on how these }
         { parameters work.  In the case of a list decoration, the width, once established }
         { in the usual way, is fixed; it will not change when the window is resized even if }
         { you specify the width with a number <= 0 in this SetUp procedure. }
         { NOTE:  The font, type size, and font style used for the entries in the list will }
         { be those in effect in the window when the list decoration is SetUp.  To use a }
         { font different from the standard widow font, you can:  Set the port to win.theWindow, }
         { set the font data with TextSize, TextFont and TextFace, and THEN call the SetUp}
         { procedure for the list (after doing that, you can reset the font data for the window }
         { if you want. }
			procedure InsertEntry (entry: string;
										beforeNumber: integer);
         { Add a single entry to the list.  The entry will be inserted in the list before the }
         { current entry number beforeNumber;  NOTE THAT entries are numbered starting }
         { with 0.  If beforeNumber is greater than the number of the last item in the list, }
         { then the new item is added at the end of the list. }
			procedure InsertEntries (entries: string;
										beforeNumber: integer);
         { adds one or more entries to the list.  The string entries specifies the items, }
         { separated by backslashes.  It is not possible to add an empty entry or an entry }
         { containing a backslash with this procedure (you can do so with InsertEntry). }
         { The parameter beforeNumber has the same meaning as in procedure InsertEntry. }
         { Use this procedure when you want to insert a number of entries at once; it }
         { just redraws the list once, at the end.  InsertEntry will redraw it after each }
         { insertion. }
			procedure InsertEntryAlphabetically (entry: string);
         { Inserts the entry before the first entry in the list that follows the parameter }
         { entry in alphabetically order.  This will not alphabetize the list, but if it is }
         { already in alphabetical order, inserting a new entry with this procedure will }
         { preserve that order. }
			procedure HandleDoubleClick;
         { This procedure is called when the user double-clicks on an entry in the list }
         { The default entry does nothing; you can override this method if you want to }
         { respond to a double-click.  }
			function FindEntry (entry: string): integer;
         { Searches the list for specified entry; if found, it returns the position of the }
         { entry in the list (>= 0); if it is not found, a value of  -1  is returned. }
			procedure DeleteEntry (entry: string);
         { If the specified entry is in the list, it is deleted }
			procedure DeleteEntryNumber (n: integer);
         { deletes entry number n; this will have no effect if the list does not have an }
         { n-th element. }
			procedure DeleteAllEntries;
         { quickly clearts all items from the list. }
			procedure GetEntryNumber (num: integer;
										var entry: string);
         { returns the num-th entry in the list; returns the empty string if num is < 0 or }
         { is greater than the number of the last item in the list. }
			function SelectedEntryNumber: integer;
         { Returns the number of the currently selected entry; returns -1 if no entry }
         { is currently selected. }
			procedure GetSelectedEntry (var entry: string);
         { Returns a copy of the currently selected entry; returns the empty string if }
         { no item is currently selected. }
			procedure SelectEntryNumber (num: integer);
         { Selects the specified entry number; if num = -1 (or in general if it is not }
         { the number of an item in the list) then the empty string is returned. }
			procedure DisplayEntriesOnly (displayOnly: boolean);
         { If displayOnly is true, then the user will not be able to select an item by }
         { clicking on it, and double-clicking will also have no effect.  The currently }
         { selected item, if any, will also be de-hilited.  Calling this procedure with }
         { displayOnly = false will restore the default behaviour. }
			procedure adjustSize;
			override;
         { modified to resize the list when the decoration is resized }
			procedure SetSize (newWidth, newHeight: integer);
			override;
         { modified so that the newWidth specified is ignored }
			procedure doActivate (active: boolean);
			override;
         { modified to activate/deactivate the list }
			procedure doDraw;
			override;
         { modified to draw the list }
			procedure doClick (localPt: point;
										modifiers: longint);
			override;
         { modified to handle a user click on the list or its scroll bar }
			procedure kill;
			override;
         { modified to dispose of the list data structure }
		end;


implementation

type
	stringList = array[1..20] of string;

procedure xList.SetUp (win: xWindow;
								initialEntries: string;
								theLeft, theRight, fixedWidth, theHeight: integer);
	var
		size: point;
		listRect, bounds: Rect;
	begin
		init;
		theList := nil;
		Install(win, theLeft, theRight, fixedWidth, theHeight);
		width := drawRect.right - drawRect.left;   { make width be fixed }
		listRect := clickRect;
		listRect.right := listRect.right - 15;
		SetRect(bounds, 0, 0, 1, 0);
		size.h := listRect.right - listRect.left;
		size.v := 0;
		theList := LNew(listRect, bounds, size, 0, itsWindow.theWindow, false, false, false, true);
		theList^^.SelFlags := lOnlyOne;
		if initialEntries <> '' then
			InsertEntries(initialEntries, -1);
		wantsClick := true;
		noUserSelect := false;
		saveFont := itsWindow.theWindow^.txFont;
		saveSize := itsWindow.theWindow^.txSize;
		saveFace := itsWindow.theWindow^.txFace;
	end;

procedure xList.InsertEntry (entry: string;
								beforeNumber: integer);
	var
		theCell: point;
		oldFont, oldSize: integer;
		oldFace: style;
		savePort: GrafPtr;
	begin
		GetPort(savePort);
		SetPort(itsWindow.theWindow);
		if beforeNumber < 0 then
			beforeNumber := 0;
		oldFont := itsWindow.theWindow^.txFont;
		oldSize := itsWindow.theWindow^.txSize;
		oldFace := itsWindow.theWindow^.txFace;
		TextFont(saveFont);
		TextSize(saveSize);
		TextFace(saveFace);
		beforeNumber := LAddRow(1, beforeNumber, theList);
		theCell.h := 0;
		theCell.v := beforeNumber;
		LSetCell(@entry[1], length(entry), theCell, theList);
		TextFont(oldFont);
		TextSize(oldSize);
		TextFace(oldFace);
		SetPort(savePort);
	end;

procedure SubdivideString (var str: string;
								var nameList: stringList;
								var ct: integer);
	var
		i: integer;
	begin
		if length(str) = 0 then begin
				ct := 0;
				EXIT(SubdivideString);
			end;
		for i := 1 to 20 do
			nameList[i] := '';
		ct := 1;
		i := 1;
		while i <= length(str) do begin
				if str[i] = '\' then begin
						if ct = 20 then
							i := length(str)
						else if nameList[ct] <> '' then
							ct := ct + 1;
					end
				else
					nameList[ct] := Concat(nameList[ct], str[i]);
				i := i + 1;
			end;
		if nameList[ct] = '' then
			ct := ct - 1;
	end;

procedure xList.InsertEntries (entries: string;
								beforeNumber: integer);
	var
		entry: stringList;
		ct: integer;
		loc: integer;
		theCell: point;
		i: integer;
	begin
		SubdivideString(entries, entry, ct);
		if ct = 0 then
			EXIT(InsertEntries);
		LDoDraw(false, theList);
		if beforeNumber < 0 then
			beforeNumber := 0;
		beforeNumber := LAddRow(ct, beforeNumber, theList);
		theCell.h := 0;
		for i := 1 to ct do begin
				theCell.v := beforeNumber + i - 1;
				LSetCell(@entry[i, 1], length(entry[i]), theCell, theList);
			end;
		forceRedraw;
		LDoDraw(true, theList);
	end;

procedure xList.InsertEntryAlphabetically (entry: string);
	var
		last, loc: integer;
		theCell: Cell;
		str: string;
		done: boolean;
		oldFont, oldSize: integer;
		oldFace: style;
		saveport: GrafPtr;
	begin
		last := theList^^.dataBounds.bottom;
		loc := 0;
		done := false;
		repeat
			if loc = last then
				done := true
			else begin
					GetEntryNumber(loc, str);
					if RelString(entry, str, false, false) = -1 then
						done := true
					else
						loc := loc + 1
				end
		until done;
		GetPort(savePort);
		Setport(itsWindow.theWindow);
		oldFont := itsWindow.theWindow^.txFont;
		oldSize := itsWindow.theWindow^.txSize;
		oldFace := itsWindow.theWindow^.txFace;
		TextFont(saveFont);
		TextSize(saveSize);
		TextFace(saveFace);
		loc := LAddRow(1, loc, theList);
		theCell.h := 0;
		theCell.v := loc;
		LSetCell(@entry[1], length(entry), theCell, theList);
		TextFont(oldFont);
		TextSize(oldSize);
		TextFace(oldFace);
		SetPort(savePort);
	end;

function xList.FindEntry (entry: string): integer;
	var
		i: integer;
		name: string;
	begin
		for i := 0 to theList^^.databounds.bottom - 1 do begin
				GetEntryNumber(i, name);
				if name = entry then begin
						FindEntry := i;
						Exit(FindEntry)
					end
			end;
		FindEntry := -1;
	end;

procedure xList.DeleteEntry (entry: string);
	var
		n: integer;
	begin
		n := FindEntry(entry);
		DeleteEntryNumber(n);
	end;

procedure xList.DeleteEntryNumber (n: integer);
	var
		oldFont, oldSize: integer;
		oldFace: style;
		savePort: GrafPtr;
	begin
		if n >= 0 then begin
				GetPort(savePort);
				SetPort(itsWindow.theWindow);
				oldFont := itsWindow.theWindow^.txFont;
				oldSize := itsWindow.theWindow^.txSize;
				oldFace := itsWindow.theWindow^.txFace;
				TextFont(saveFont);
				TextSize(saveSize);
				TextFace(saveFace);
				LDelRow(1, n, theList);
				TextFont(oldFont);
				TextSize(oldSize);
				TextFace(oldFace);
				SetPort(savePort);
			end;
	end;

procedure xList.DeleteAllEntries;
	begin
		LDelRow(0, 1, theList);
	end;

procedure xList.GetEntryNumber (num: integer;
								var entry: string);
	var
		text: string;
		charct: integer;
		theCell: Cell;
	begin
		theCell.h := 0;
		theCell.v := num;
		charct := 255;
   {$PUSH}
   {$R-}
		LGetCell(@text[1], charct, theCell, theList);
		text[0] := chr(charct);
		entry := text;
   {$POP}
	end;

function xList.SelectedEntryNumber: integer;
	var
		theCell: cell;
	begin
		theCell.h := 0;
		theCell.v := 0;
		if LGetSelect(true, theCell, theList) then
			SelectedEntryNumber := theCell.v
		else
			SelectedEntryNumber := -1
	end;

procedure xList.GetSelectedEntry (var entry: string);
	var
		current: integer;
	begin
		current := SelectedEntryNumber;
		GetEntryNumber(current, entry)
	end;

procedure xList.SelectEntryNumber (num: integer);
	var
		theCell: cell;
		current: integer;
	begin
		current := SelectedEntryNumber;
		if current <> num then begin
				theCell.h := 0;
				if current <> -1 then begin
						theCell.v := current;
						LSetSelect(false, theCell, theList);
					end;
				if num >= 0 then begin
						theCell.v := num;
						LSetSelect(true, theCell, theList);
					end;
			end
	end;

procedure xList.HandleDoubleClick;
	begin
		SysBeep(5);
	end;

procedure xList.DisplayEntriesOnly (displayOnly: boolean);
	begin
		noUserSelect := displayOnly;
		if displayOnly then
			SelectEntryNumber(-1);
	end;

procedure xList.adjustSize;
	var
		w, h: integer;
		R: Rect;
	begin
		inherited adjustsize;
		insetRect(clickRect, 1, 1);
		if theList <> nil then begin
				R := clickRect;
				R.right := R.right - 15;
				theList^^.rView := R;
				w := clickRect.right - clickRect.left - 15;
				h := clickRect.bottom - clickRect.top;
				LDoDraw(false, theList);
				LSize(w, h, theList);
				LDoDraw(true, theList);
			end;
	end;

procedure xList.SetSize (newWidth, newHeight: integer);
	begin
		inherited SetSize(width, newHeight);  { change in width not allowed }
	end;

procedure xList.doActivate (active: boolean);
	var
		savePort: GrafPtr;
		oldFont, oldSize: integer;
		oldFace: style;
	begin
		GetPort(savePort);
		setPort(itsWindow.theWindow);
		oldFont := itsWindow.theWindow^.txFont;
		oldSize := itsWindow.theWindow^.txSize;
		oldFace := itsWindow.theWindow^.txFace;
		TextFont(saveFont);
		TextSize(saveSize);
		TextFace(saveFace);
		LActivate(active, theList);
		TextFont(oldFont);
		TextSize(oldSize);
		TextFace(oldFace);
		SetPort(savePort);
	end;

procedure xList.doDraw;
	var
		rgn: RgnHandle;
		savePen: point;
		oldFont, oldSize: integer;
		oldFace: style;
	begin
		savePen := itsWindow.theWindow^.pnSize;
		PenSize(1, 1);
		FrameRect(drawRect);
		PenSize(savePen.h, savePen.v);
		rgn := itsWindow.theWindow^.visRgn;
		oldFont := itsWindow.theWindow^.txFont;
		oldSize := itsWindow.theWindow^.txSize;
		oldFace := itsWindow.theWindow^.txFace;
		TextFont(saveFont);
		TextSize(saveSize);
		TextFace(saveFace);
		LUpdate(rgn, theList);
		TextFont(oldFont);
		TextSize(oldSize);
		TextFace(oldFace);
	end;

procedure xList.doClick (localPt: point;
								modifiers: longint);
	var
		dbl: boolean;
		oldFont, oldSize: integer;
		oldFace: style;
	begin
		if (localPt.h > clickRect.right - 15) then begin
				oldFont := itsWindow.theWindow^.txFont;
				oldSize := itsWindow.theWindow^.txSize;
				oldFace := itsWindow.theWindow^.txFace;
				TextFont(saveFont);
				TextSize(saveSize);
				TextFace(saveFace);
				dbl := LClick(localPt, modifiers, theList);   { scroll }
				TextFont(oldFont);
				TextSize(oldSize);
				TextFace(oldFace);
			end
		else if not noUserSelect then begin
				dbl := LClick(localPt, modifiers, theList);
				if dbl then
					HandleDoubleClick;
			end;
	end;

procedure xList.kill;
	begin
		LDispose(theList);
		inherited kill;
	end;

end.